local fps = 1/40;
local function Engine_OnKeyDown(self, caller, key)
	if not self.protagonist then return end
	if key == "P" then
		if self.running then
			self.running = false;
			self.backdrop:SetTexture(0.2,0.2,0.3,0.4);
			self.caption:SetText("Paused!");
		else
			self.running = true;
			self.backdrop:SetTexture(nil);
			self.caption:SetText("");
		end
	end
	if not self.running then return end
	if 		key == "LEFT" then
		self.protagonist.direction = "left"
		self.keys.leftDown = true;
		self.keys.rightDown = false;
	elseif	key == "RIGHT" then
		self.protagonist.direction = "right"
		self.keys.leftDown = false;
		self.keys.rightDown = true;
	elseif key == "UP" then
		self.keys.jumpFired = true;
	elseif key == "SPACE" then
		self.keys.spaceFired = true;
	end
end

local function Engine_OnKeyUp(self, caller, key)
	if not self.protagonist then return end
	if 		key == "LEFT" then
		self.keys.leftDown = false;
	elseif	key == "RIGHT" then
		self.keys.rightDown = false;
	elseif key == "SPACE" then
		self.keys.spaceFired = false;
	elseif key == "UP" then
		self.keys.jumpFired = false;
	end
end

local function Engine_SetSpeed(self, speed)
	if not self.running then return end
	self.gameSpeed = tonumber(speed) or self.gameSpeed;
end

local function Engine_Update(self, ...)
	local speed = 0;
	if not self.running then return end
	if self.keys.leftDown == true then speed = -(self.protagonist.speed); end
	if self.keys.rightDown == true then speed = (self.protagonist.speed); end
	self.protagonist.forceH = speed;
	if self.keys.jumpFired and self.protagonist:OnGround() then
		self.protagonist:Jump()
		self.keys.jumpFired = false;
	end

	local now = GetTime();
	if (now - self.lastUpdate) >= (fps) then
		self.world:Step(1/self.gameSpeed);
		self.lastUpdate = now;
	end
end

local function Engine_CallBack(t, mob1, side1, mob2, side2)
	if mob1.static == true then
		return "BLOCK", mob2
	elseif mob2.static == true then
		return "BLOCK", mob1
	end

end

local function Engine_LoadWorld(self, ...)
	if not self.world then
		self.world = Gnomario_World_Create(self.frame,480,320, self.CallBack);
	end
	self.world:Clear();
end

local function Engine_Start(self)
	self.running = true;
	self.lastUpdate = GetTime();
end

local function Engine_SetProtagonist(self, mob)
	self.protagonist = mob;
end

function Gnomario_Engine(parent)
	local engine = {};
	engine.sprites = {};
	engine.keys = {};
	engine.frame = CreateFrame("Frame");
	engine.frame:SetParent(parent);
	engine.frame:EnableKeyboard(true);
	engine.frame:SetAllPoints(parent);
	engine.frame:SetScript("OnKeyDown", function(...) Engine_OnKeyDown(engine,...) end);
	engine.frame:SetScript("OnKeyUp", function(...) Engine_OnKeyUp(engine,...) end);
	engine.frame:SetScript("OnUpdate", function(...) Engine_Update(engine, ...) end);
	engine.backdrop = engine.frame:CreateTexture("OVERLAY");
	engine.backdrop:SetAllPoints(engine.frame);
	-- caption
	local caption = engine.frame:CreateFontString(nil, "DIALOG");
	caption:SetParent(engine.frame);
	caption:SetFontObject(ChatFontNormal);
	caption:SetText("");
	caption:SetPoint("TOPLEFT", engine.frame, "TOPLEFT", 5, -5);
	caption:SetPoint("BOTTOMRIGHT", engine.frame, "TOPRIGHT", 5, -25);
	caption:SetTextColor(0.9,0.5,0.2,1);
	engine.caption = caption;

	engine.gameSpeed = 25;
	engine.running = false;
	engine.SetSpeed = Engine_SetSpeed;
	engine.LoadWorld = Engine_LoadWorld;
	engine.Start = Engine_Start;
	engine.SetProtagonist = Engine_SetProtagonist;
	engine.CallBack = Engine_CallBack;
	return engine;
end
