local function deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

local function Gnomario_World_Unregister(self)
	local sprites = self.sprites;
	for n = 1, #sprites do
		local sprite = sprites[n];
		sprite.texture:Hide();
		sprite.texture:SetTexture(nil);
		sprite.id = nil;
	end
	self.world = {};
	self.creatures = {};
	self.environment = {};
end

local function Gnomario_World_Add(self)
	local sprites = self.sprites;
	for n = 1, #sprites do
		local sprite = sprites[n];
		if sprite and not sprite.id then
			sprite.id = time();
			return sprite;
		end
	end
	local sprite = {};
	sprite.texture = UIParent:CreateTexture();
	sprite.id = time();
	table.insert(self.sprites, sprite);
	return sprite;
end

function Gnomario_World_Spawn(self, type, db)
	db = db or Gnomario_Creatures;
	if not db then error("Could not find creature database"); end
	local mob = {};
	for key, obj in pairs(db) do
		if key == type then
			mob = deepcopy(obj);
			break;
		end
	end
	mob.width = mob.width or 20;
	mob.height = mob.height or 20;
	mob.sprite = Gnomario_World_Add(self);
	if mob.tex then
		mob.sprite.texture:SetTexture([[Interface\AddOns\Gnomario\Textures\]] .. mob.tex);
	else
		mob.sprite.texture:SetTexture(1,1,0,1);
	end
	mob.sprite.texture:SetSize(mob.width, mob.height);
	mob.sprite.texture:SetParent(self.frame);
	mob.sprite.obj = mob;
	tinsert(self.creatures, mob);

	self.physics:Put(mob, 0, 0);
	return mob;
end

function Gnomario_World_Step(self, duration)
	self.physics:Step(duration);
	for n = 1, #self.sprites do
		local sprite = self.sprites[n];
		if sprite.id then
			sprite.texture:Show();
			sprite.texture:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", sprite.obj.x, sprite.obj.y);
		else
			sprite.texture:Hide();
		end
	end
end


function Gnomario_World_Create(frame,width,height,cb)
	local world = {};
	world.frame = frame;
	world.environment = {};
	world.creatures = {};
	world.sprites = {};
	world.width = width;
	world.height = height;
	world.Clear = Gnomario_World_Unregister;
	world.Spawn = Gnomario_World_Spawn;
	world.physics = Gnomario_Physics_Chamber(width, height, cb);
	world.Step = Gnomario_World_Step;
	return world;
end
